﻿
package net.kaikoga.text {
	
	import flash.display.Sprite;
	import flash.display.Stage;
	import flash.events.Event;
	import flash.events.MouseEvent;
	import flash.events.KeyboardEvent;
	import flash.events.TextEvent;
	import flash.text.TextField;
	import flash.text.TextFieldType;
	import flash.ui.Keyboard;
	
	/**
	* フォーカスを受け取った状態で、キーボードからの入力が行われると送出される。
	* @eventType	net.kaikoga.text.TextDownEvent.TEXT_DOWN
	*/
	[Event(name="textDown", type="net.kaikoga.text.TextDownEvent")]
	/**
	* フォーカスを受け取った状態で入力が行われた後、キーが放されると送出される。
	* @eventType	net.kaikoga.text.TextDownEvent.TEXT_UP
	*/
	[Event(name="textUp", type="net.kaikoga.text.TextDownEvent")]

	/**
	* <p>マウスクリックすることにより自動的にフォーカスされる。<br />
	* 更に、フォーカスを受け取ることにより、キーボードから入力された文字や一部の特殊キーを扱うTextDownEventを送出する。<br />
	* おそらくエディタなどの作成に便利だと思う。<br />
	* TODO（現段階ではフォーカスをグラフィカルに表示する手段がないです）</p>
	*
	* <p>手動でフォーカスを与えるには、インスタンスをステージに追加した後にインスタンスのsetFocusを呼んでやって下さい。</p>
	* <listing>
	* var tds:TextDownSprite = new TextDownSprite();
	* this.addChild(tds);
	* tds.setFocus();
	* </listing>
	* @author kaikoga
	*/
	public class TextDownSprite extends Sprite {
		
		private var m_Text:TextField;//For receiving TextEvents
		
		private var keyCode:int;
		private var keyLocation:int;
		private var shiftKey:Boolean;
		private var altKey:Boolean;
		private var ctrlKey:Boolean;

		private var keyCodeToCharCode:Array/*int*/;
		
		/**
		* 改行文字の文字コード。ここの設定を変えると、入力テキストの改行が自動変換される。<br />
		* nullにすると改行文字の変換が行われなくなるが、文字列をCtrl+Vでペーストした時の挙動がOS依存になる。
		* @default 改行文字（文字コードは16進0A）。
		*/
		public var lineFeed:String = "\x0a";

		/**
		* trueなら、Ctrl+Vでペーストした文字列やWindowsの改行コードは文字ごとにイベントが発生する。<br />
		* falseなら、Ctrl+Vでペーストした文字列やWindowsの改行コードに対するイベントの中身は不定。
		* @default true
		*/
		public var splitStrings:Boolean = true;
		
		/**
		 * コンストラクタ。
		 * @param	autofocus trueなら、ステージに追加されたときに自動的にフォーカスを受け取る。
		 * @param	proxy trueなら、当たり判定が自動生成される。trueにすると、クリック可能な子供をaddChildしたりhitAreaを手動で設定したりしなくてもクリック可能になる。
		 * @param	width proxy=trueの場合の、当たり判定の幅。
		 * @param	height proxy=trueの場合の、当たり判定の高さ。
		 */
		public function TextDownSprite(autofocus:Boolean = false, proxy:Boolean = false, width:int = 100, height:int = 100) {
			super();
			this.keyCodeToCharCode = [];
			this.m_Text = new TextField();
			this.m_Text.width = 0;
			this.m_Text.height = 0;
			this.m_Text.type = TextFieldType.INPUT;
			this.m_Text.multiline = true;
			this.m_Text.visible = false;
			this.m_Text.addEventListener(TextEvent.TEXT_INPUT, this.onTextInput);
			this.m_Text.addEventListener(KeyboardEvent.KEY_DOWN, this.onKeyDown);
			this.m_Text.addEventListener(KeyboardEvent.KEY_UP, this.onKeyUp);
			this.addEventListener(MouseEvent.MOUSE_DOWN, this.onMouseDown);
			if (autofocus) {
				this.addEventListener(Event.ADDED_TO_STAGE, this.onAddedToStage);
			}
			if (proxy) {
				var spr:Sprite = new Sprite();
				spr.graphics.beginFill(0);
				spr.graphics.drawRect(0, 0, width, height);
				this.addChild(spr);
				spr.visible = false;
				this.hitArea = spr;
			}
			this.addChild(this.m_Text);
		}
		
		private function onAddedToStage(event:Event):void {
			this.setFocus();
		}

		private function onMouseDown(event:MouseEvent):void {
			this.setFocus();
		}

		/**
		 * フォーカスを設定する。※TextDownSpriteのインスタンスを直接stage.focusに設定してはダメです
		 */
		public function setFocus(stage:Stage = null):void {
			if (!stage) {
				stage = this.stage;
			}
			if (stage) {
				stage.focus = this.m_Text;
			}
		}
		
		private function dispatchTextDownEvent(charCode:int):void {
			this.dispatchEvent(new TextDownEvent(TextDownEvent.TEXT_DOWN, false, true, charCode,
					this.keyCode, this.keyLocation,
					this.altKey, this.ctrlKey, this.shiftKey));
			this.keyCodeToCharCode[keyCode] = charCode;
		}
		
		// イベントが呼び出される順番：(keyDown→textInput)xn→keyUp
		// キーが押された際は、keyDownでキーボードイベントの値を記録してtextInputでイベントを送出する。
		// キーが離された際は、textInputでテキストイベントの値を記録してkeyUpでイベントを送出する。
		// keyUpの前にkeyDownが複数回発生する可能性がある。
		
		private function onTextInput(event:TextEvent):void {
			event.preventDefault();
			var text:String = event.text;
			if (this.lineFeed) {
				text = text.replace(/\x0d\x0a|\x0d|\x0a/g, this.lineFeed);
			}
			var len:int = text.length;
			if (splitStrings && (len != 1)) {
				for (var i:int = 0; i < len; i++) {
					this.dispatchTextDownEvent(text.charCodeAt(i));
				}
			} else {
				this.dispatchTextDownEvent(text.charCodeAt());
			}
		}
		
		private function onKeyDown(event:KeyboardEvent):void {
			this.keyCode = event.keyCode;
			this.keyLocation = event.keyLocation;
			this.shiftKey = event.shiftKey;
			this.ctrlKey = event.ctrlKey;
			this.altKey = event.altKey;
			//catch control inputs here
			switch (event.keyCode) {
				case Keyboard.BACKSPACE:
				this.dispatchTextDownEvent(TextDownEvent.BACKSPACE);
				break;
				case Keyboard.DELETE:
				this.dispatchTextDownEvent(TextDownEvent.DELETE);
				break;
				case Keyboard.TAB:
				this.dispatchTextDownEvent(TextDownEvent.TAB);
				break;
				case Keyboard.ESCAPE:
				this.dispatchTextDownEvent(TextDownEvent.ESCAPE);
				break;
				case Keyboard.UP:
				this.dispatchTextDownEvent(TextDownEvent.UP);
				break;
				case Keyboard.DOWN:
				this.dispatchTextDownEvent(TextDownEvent.DOWN);
				break;
				case Keyboard.RIGHT:
				this.dispatchTextDownEvent(TextDownEvent.RIGHT);
				break;
				case Keyboard.LEFT:
				this.dispatchTextDownEvent(TextDownEvent.LEFT);
				break;
			}
		}
		
		public function onKeyUp(event:KeyboardEvent):void {
			if (this.willTrigger(TextDownEvent.TEXT_UP)) {
				var charCode:int = this.keyCodeToCharCode[event.keyCode];
				if (charCode) {
					this.dispatchEvent(new TextDownEvent(TextDownEvent.TEXT_UP, false, true, charCode,
							event.keyCode, event.keyLocation,
							event.altKey, event.ctrlKey, event.shiftKey));
				}
			}
		}
		
	}
	
}
